__global__ void real2complex ( cufftDoubleReal *out, double *in, int P, int M, int rowsize, int bias1, int bias2, int bias3, int bias4 )
{
	int t = blockIdx.x * THREADS + threadIdx.x;
	if (IsValidRange)
	{
		int j = (int) (t / M);
		int k = t % M;
		out[bias1 + j*rowsize + k] = out[bias4 - j*rowsize - k] = in[t];
		out[bias2 + j*rowsize - k] = out[bias3 - j*rowsize + k] = -in[t];
	}
}
__global__ void solve_poisson ( cufftDoubleReal *out, cufftDoubleComplex *in, int P, int M, int rowsize, int bias1, int bias2, int bias3, int bias4, double *kx )
{
	int t = blockIdx.x * THREADS + threadIdx.x;
	if (IsValidRange)
	{
		int j = (int) (t / M);
		int k = t % M;
		out[bias1 + j*rowsize + k] = out[bias4 - j*rowsize - k] = -0.25 * in[bias1 - M + t + 2*j].x / ( kx[j] + kx[k] );
		out[bias2 + j*rowsize - k] = out[bias3 - j*rowsize + k] = -out[bias1 + j*rowsize + k];
	}
}
__global__ void solve_poisson_pcgls ( cufftDoubleReal *out, cufftDoubleComplex *in, int P, int M, int rowsize, int bias1, int bias2, int bias3, int bias4, double *kx )
{
	int t = blockIdx.x * THREADS + threadIdx.x;
	if (IsValidRange)
	{
		int j = (int) (t / M);
		int k = t % M;
		out[bias1 + j*rowsize + k] = out[bias4 - j*rowsize - k] = -0.25 * in[bias1 - M + t + 2*j].x / sqrt( kx[j] + kx[k] );
		out[bias2 + j*rowsize - k] = out[bias3 - j*rowsize + k] = -out[bias1 + j*rowsize + k];
	}
}
__global__ void complex2real ( double *out, cufftDoubleComplex *in, int P, int M, int rowsize, int bias1, double scale )
{
	int t = blockIdx.x * THREADS + threadIdx.x;
	if (IsValidRange)
	{
		int j = (int) (t / M);
		out[t] = scale * in[bias1 - M + t + 2*j].x;
	}
}

__global__ void real2complex_dft ( cufftDoubleComplex *out, double *in, int P, int M )
{
	int t = blockIdx.x * THREADS + threadIdx.x;
	if (IsValidRange)
	{
		out[t].x = in[t] / M;
		//out[t].y = 0.0;
	}
}
__global__ void complex_shuffle ( cufftDoubleComplex *r, int M )
{
	int a = threadIdx.x + 1, b = M - threadIdx.x - 1, P = M*M;
	double tmp;
	for (int i=0; i<P; i+=M)
	{
		tmp = r[a+i].x;
		r[a+i].x = r[b+i].x;
		r[b+i].x = tmp;
		tmp = r[a+i].y;
		r[a+i].y = r[b+i].y;
		r[b+i].y = tmp;
	}
	a = M * a;
	b = M * b;
	for (int i=0; i<M; i++)
	{
		tmp = r[a+i].x;
		r[a+i].x = r[b+i].x;
		r[b+i].x = tmp;
		tmp = r[a+i].y;
		r[a+i].y = r[b+i].y;
		r[b+i].y = tmp;
	}
}
__global__ void solve_poisson_dft ( cufftDoubleComplex *out, cufftDoubleComplex *in, int P, int M, double *kx )
{
	int t = blockIdx.x * THREADS + threadIdx.x;
	if (IsValidRange)
	{
		int j = (int) (t / M);
		int k = t % M;
		//out[t].x = in[t].x / ( kx[j] + kx[k] );
		//out[t].y = in[t].y / ( kx[j] + kx[k] );
		/*
		//shuffling index
		int l;
		if (j > 0)
		{
			if (k > 0)
				l = M*M - j*M - k + M;
			else
				l = M*M - j*M;
		}
		else
		{
			if (k > 0)
				l = M - k;
			else
				l = 0;
		}
		*/
		out[t].x = in[t].x / ( kx[j] + kx[k] );
		out[t].y = in[t].y / ( kx[j] + kx[k] );
	}
}
__global__ void solve_poisson_pcgls_dft ( cufftDoubleComplex *in, int P, int M, double *kx )
{
	int t = blockIdx.x * THREADS + threadIdx.x;
	if (IsValidRange)
	{
		int j = (int) (t / M);
		int k = t % M;
		/*
		//shuffling index
		int l;
		if (j > 0)
		{
			if (k > 0)
				l = M*M - j*M - k + M;
			else
				l = M*M - j*M;
		}
		else
		{
			if (k > 0)
				l = M - k;
			else
				l = 0;
		}
		*/
		in[t].x = in[t].x / sqrt( kx[j] + kx[k] );
		in[t].y = in[t].y / sqrt( kx[j] + kx[k] );
	}
}
__global__ void complex2real_dft ( double *out, cufftDoubleComplex *in, int P, int M )
{
	int t = blockIdx.x * THREADS + threadIdx.x;
	if (IsValidRange)
	{
		out[t] = in[t].x / M;
	}
}